{ **************************************************************************** }
{ **    Addin Development Kit                                               ** }
{ **    (c) 2000 by Dott. Cosimo De Michele                                 ** }
{ **************************************************************************** }

unit ADKMng;

interface

uses
  windows,
  forms,
  classes;

const
  ALL_ADDINS = -1;

type
  TConnectMode = (cmAfterStartup, cmStartup, cmExternal);
  TDisconnectMode = (dmHostShutdown, dmUserClosed);

  TConnectionEvent = procedure(Sender : TObject; ConnectMode: TConnectMode; var AllowConnection: boolean; var custom: variant) of object;
  TDisconnectionEvent = procedure(Sender : TObject; RemoveMode: TDisconnectMode; var custom: variant) of object;
  TAddinMessageEvent = procedure(Sender : TObject; AddinMessage : integer; var MessageParam : variant) of object;
  TAddinEvent = procedure(Sender : TObject; var custom: variant) of object;

  TAddIn = class(TDataModule)
  private
    FAuthor,
    FCopyright,
    FVersion: string;
    FInstanceCount: integer;
    FConnected: Boolean;
    FModuleInstance: HMODULE;
    { private events }
    FOnAddinMessage: TAddinMessageEvent;
    FOnConnection: TConnectionEvent;
    FOnDisconnection: TDisconnectionEvent;
    FOnRegister: TAddinEvent;
    { private property methods }
    function GetDescription: string;
    function GetVersion: string;
    function GetIndex: integer;
    procedure SetVersion(const value: string);
    function GetPackageName: string;
  public
    constructor CreateModule(AModuleInstance: HMODULE);
    destructor Destroy; override;
    property PackageName: string read GetPackageName;
    property Description : string read GetDescription;
    property Index: integer read GetIndex;
    property ModuleInstance: HMODULE read FModuleInstance;
    property InstanceCount : integer read FInstanceCount;
    { methods }
    function Connect(ConnectMode: TConnectMode; var custom: variant): boolean; virtual;
    procedure Disconnect(RemoveMode: TDisconnectMode; var custom: variant); virtual;
    procedure SendMessage(AddinMessage : integer; var MessageParam : variant); virtual;
    procedure Register(var custom: variant); virtual;
  published
    property Author : string read FAuthor write FAuthor;
    property Copyright : string read FCopyright write FCopyright;
    property Version : string read GetVersion write SetVersion;
    { events }
    property OnConnection: TConnectionEvent read FOnConnection write FOnConnection;
    property OnDisconnect: TDisconnectionEvent read FOnDisconnection write FOnDisconnection;
    property OnAddinMessage: TAddinMessageEvent read FOnAddinMessage write FOnAddinMessage;
    property OnRegister: TAddinEvent read FOnRegister write FOnRegister;
  end;

  TAddinClass = class of TAddin;

  function InitAddin(AModule: HMODULE; AddinClass: TAddinClass): TAddin;
  function AddinsCount: integer;
  function GetAddinByIndex(index: integer): TAddin;
  function GetAddinByModule(AModule: HMODULE): TAddin;
  function GetAddinByName(PackageName: string): TAddin;
  function LoadAddin(const PackageName: string): TAddin;
  procedure UnloadAddin(const PackageName: string);
  procedure LoadAddins(const PackageFolder, PackageExtension: string);
  procedure UnloadAddins;
  procedure SendAddinMessage(AddinIndex: integer; AddinMessage: integer;
                             var MessageParam: variant);
  //procedure LoadRegisteredAddins();
  //procedure LoadRegisteredAddin()
  //procedure RegisterAddin();
  //procedure UnregisterAddin();

implementation

{$R *.DFM}

uses
  //forms,
  dialogs,
  sysutils;

var
  FLoadedAddins: TList;

{ **************************************************************************** }
{ **                          G e n e r a l                                 ** }
{ **************************************************************************** }

{ Return the description of a loaded module }
function GetLoadedPackageDescription(AModule: HMODULE): string;
var
  ResModule: HModule;
  ResInfo: HRSRC;
  ResData: HGLOBAL;
begin
  Result := '';
  if AModule = 0 then
    Exit;
  ResModule := FindResourceHInstance(AModule);
  try
    ResInfo := FindResource(ResModule, 'DESCRIPTION', RT_RCDATA);
    if ResInfo <> 0 then
    begin
      ResData := LoadResource(ResModule, ResInfo);
      if ResData <> 0 then
      try
        Result := PWideChar(LockResource(ResData));
        UnlockResource(ResData);
      finally
        FreeResource(ResData);
      end;
    end;
  finally
    //FreeLibrary(ResModule);
  end;
end;

procedure UnloadPackageModule(AModule: HMODULE);
var
  i: Integer;
begin
  { Make sure there aren't any instances of any of the classes from Module
    instantiated, if so then free them.  (this assumes that the classes are
    owned by the application) }
  with Application do
    for i := ComponentCount - 1 downto 0 do
      if HMODULE(FindClassHInstance(Components[i].ClassType)) = AModule then
        Components[i].Free;
  UnRegisterModuleClasses(AModule);
  UnLoadPackage(AModule);
end;

{returns the loaded module, creating it if not found}
function InitAddin(AModule: HMODULE; AddinClass: TAddinClass): TAddin;
begin
  Result := GetAddinByModule(AModule);
  if (Result = nil) and (AddinClass<>nil) then
      Result := AddinClass.CreateModule(AModule);
end;

{ **************************************************************************** }
{ **                     T A d d I n M a n a g e r                          ** }
{ **************************************************************************** }

function AddinsCount: integer;
begin
  Result := FLoadedAddins.Count;
end;

function GetAddinByIndex(index: integer): TAddin;
begin
  Result := nil;
  if index<FLoadedAddins.Count then
    Result := FLoadedAddins[index];
end;

{returns the Addin of a module}
function GetAddinByModule(AModule: HMODULE): TAddin;
var
  i: integer;
begin
  Result := nil;
  for i := FLoadedAddins.Count - 1 downto 0 do
    if TAddin(FLoadedAddins[i]).ModuleInstance = AModule then
    begin
      Result := TAddin(FLoadedAddins[i]);
      Exit;
    end;
end;

{returns the Addin of a packagename}
function GetAddinByName(PackageName: string): TAddin;
var
  i: integer;
begin
  Result := nil;
  for i := FLoadedAddins.Count - 1 downto 0 do
    if CompareText(TAddin(FLoadedAddins[i]).PackageName, PackageName)=0 then
    begin
      Result := TAddin(FLoadedAddins[i]);
      Exit;
    end;
end;

{load the package 'packagename'.}
function LoadAddin(const PackageName: string): TAddin;
var
  ThisModule: HMODULE;
begin
  Result := GetAddinByName(PackageName);
  if not Assigned(Result) then
  begin
    ThisModule := LoadPackage(PackageName);
    Result := InitAddin(ThisModule, nil);
  end;
  if Assigned(Result) then
  begin
    //inc(Result.FInstanceCount);
  end;
end;

{unload the package 'PackageName' }
procedure UnloadAddin(const PackageName: string);
var
  i: integer;
  ThisModule: HMODULE;
begin
  for i := FLoadedAddins.Count - 1 downto 0 do
    if CompareText(TAddin(FLoadedAddins[i]).PackageName, PackageName)=0 then
    begin
      ThisModule := TAddin(FLoadedAddins[i]).ModuleInstance;
      //dec(TAddin(FLoadedAddins[i]).FInstanceCount);
      if TAddin(FLoadedAddins[i]).FInstanceCount=0 then
      begin
        TAddin(FLoadedAddins[i]).free;
        UnloadPackageModule(ThisModule);
      end;
      break;
    end;
end;

{load all packages previously registered.
 call from main form OnShow, after menus have been created.}
procedure LoadAddins(const PackageFolder, PackageExtension: string);
var
  i, Found: integer;
  sr: TSearchRec;
  Path: string;
  custom: variant;
begin
   { initialing startup }
   Path := IncludeTrailingBackslash(PackageFolder);
   try
      Found := FindFirst(Path + '*.' + PackageExtension, 0, sr);
      while Found = 0 do
      begin
         LoadAddin(Path + sr.Name);
         Found := FindNext(sr);
      end;
   finally
      FindClose(sr);
   end;
   { startup complete }
   for i:=0 to (FLoadedAddins.Count - 1) do
   begin
     custom := NULL;
     if not TAddin(FLoadedAddins[i]).Connect(cmStartup, custom) then
       UnloadAddin(TAddin(FLoadedAddins[i]).PackageName);
   end;
end;

{unload all packages previously registered}
procedure UnloadAddins;
var
  i: integer;
  ThisModule: HMODULE;
  custom: variant;
begin
  { before shutdown }
  for i := FLoadedAddins.Count - 1 downto 0 do
  begin
    custom := NULL;
    TAddin(FLoadedAddins[i]).Disconnect(dmHostShutdown, custom);
  end;
  { shutdown }
  for i := FLoadedAddins.Count - 1 downto 0 do
  begin
    ThisModule := TAddin(FLoadedAddins[i]).ModuleInstance;
    TAddin(FLoadedAddins[i]).free;
    UnloadPackageModule(ThisModule);
  end;
end;

procedure SendAddinMessage(AddinIndex: integer;
                           AddinMessage: integer;
                           var MessageParam: variant);
var
  i: integer;
begin
  if AddinIndex=ALL_ADDINS then
    for i:=0 to (FLoadedAddins.Count - 1) do
      TAddin(FLoadedAddins[i]).SendMessage(AddinMessage, MessageParam)
  else
    TAddin(FLoadedAddins[AddinIndex]).SendMessage(AddinMessage, MessageParam)
end;

{ **************************************************************************** }
{ **                           T A d d I n                                  ** }
{ **************************************************************************** }

constructor TAddin.CreateModule(AModuleInstance: HMODULE);
begin
  inherited Create(Application);
  FModuleInstance := AModuleInstance;
  FInstanceCount := 0;
  FLoadedAddins.Add(Self);
end;

destructor TAddin.Destroy;
var
  custom: variant;
begin
  custom := NULL;
  if FConnected then
    Disconnect(dmUserClosed, custom);
  FLoadedAddins.Remove(Self);
  inherited Destroy;
end;

function TAddin.GetDescription: string;
begin
  Result := GetLoadedPackageDescription(FModuleInstance);
end;

function TAddin.GetVersion: string;
begin
  Result := FVersion;
end;

function TAddin.GetIndex: integer;
begin
  Result := FLoadedAddins.IndexOf(Self);
end;

procedure TAddin.SetVersion(const value: string);
begin
  if FVersion<>value then
    FVersion := value;
end;

function TAddin.GetPackageName: string;
begin
  SetLength(Result, 200);
  GetModuleFileName(FModuleInstance, pChar(Result), 200);
  Result := pChar(Result);
end;

function TAddin.Connect(ConnectMode: TConnectMode; var custom: variant): boolean;
begin
  Result := true;
  if Assigned(FOnConnection) then
    FOnConnection(Self, ConnectMode, Result, custom);
  FConnected := Result;
end;

procedure TAddin.Disconnect(RemoveMode: TDisconnectMode; var custom: variant);
begin
  if Assigned(FOnDisconnection) then
    FOnDisconnection(Self, RemoveMode, custom);
  FConnected := false;
end;

procedure TAddin.SendMessage(AddinMessage : integer; var MessageParam : variant);
begin
  if Assigned(FOnAddinMessage) then
    FOnAddinMessage(Self, AddinMessage, MessageParam);
end;

procedure TAddin.Register(var custom : variant);
begin
  if Assigned(FOnRegister) then
    FOnRegister(Self, custom);
end;

initialization
  FLoadedAddins := TList.Create;

finalization
  FLoadedAddins.Free;

end.

